﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace Hyperbolic.Vectors
{
  /// <summary> The ScreenVector class implements the coordinates of a point in the Screen space.
  /// As the screen space is represented with finite pixels, we just use int instead of float or double.
  /// </summary>
  public class ScreenVector
  {
    #region Private fields

    private static readonly double ZONE_LENGTH = 4.0; // size of the zone
    private double _x = 0; // x coord
    private double _y = 0; // y coord

    #endregion

    #region ctor

    /// <summary> Constructor.
    /// x = 0 and y = 0.
    /// </summary>
    public ScreenVector() { }

    /// <summary> Constructor copying the given screen point.
    /// </summary>
    /// <param name="z">The screen point to copy.</param>
    public ScreenVector(ScreenVector z)
    {
      _x = z._x;
      _y = z._y;
    }

    public ScreenVector(Point p)
    {
        _x = p.X;
        _y = p.Y;
    }

    /// <summary> Constructor fixing x and y.
    /// </summary>
    /// <param name="x">The x coord.</param>
    /// <param name="y">The y coord.</param>
    public ScreenVector(double x, double y)
    {
      _x = x;
      _y = y;
    }

    #endregion

    #region Public properties

    /// <summary> Gets the x coordinate.
    /// </summary>
    public double X
    {
      get { return _x; }
      set { _x = value; }
    }

    /// <summary> Gets the y coordinate.
    /// </summary>
    public double Y
    {
      get { return _y; }
      set { _y = value; }
    }

    #endregion

    #region Projection

    /// <summary> Projects the given Euclidian point on the screen plane.
    /// </summary>
    /// <param name="ze">The euclidian point.</param>
    /// <param name="sOrigin">The origin of the screen plane.</param>
    /// <param name="sMax">The (xMax, yMax) point in the screen plane.</param>
    public void ProjectionEtoS(EuclidianVector ze, ScreenVector origin, ScreenVector max)
    {
      _x = Math.Round(ze.X * max.X) + origin.X;
      _y = -Math.Round(ze.Y * max.Y) + origin.Y;
    }

    #endregion

    #region Zone containing 

    /// <summary> Is the given HtCoordS within the zone length of this HtCoordS ?
    /// </summary>
    /// <param name="zs"></param>
    /// <returns>true if it is, false otherwise</returns>
    public bool Contains(ScreenVector zs)
    {
      double __length = this.GetDistance(zs);
      return __length <= ZONE_LENGTH;
    }

    #endregion

    #region Distance

    /// <summary> Returns the distance between this point and the given point.
    /// </summary>
    /// <param name="z">The given point.</param>
    /// <returns>The distance.</returns>
    public double GetDistance(ScreenVector z)
    {
      double __d2 = (z._x - _x) * (z._x - _x) + (z._y - _y) * (z._y - _y);
      return Math.Round(Math.Sqrt(__d2));
    }

    #endregion

    #region ToString

    public override string ToString()
    {
      return "(" + _x + " : " + _y + ")S";
    }

    #endregion

  }
}
